/*
 * multiwin.c
 * - A quick and dirty example of how to make
 *   a multiple windowed SDL app.
 *   This example uses fork() and pipe()
 *
 * Samuel Hart - Code is placed in Public Domain
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#include <fcntl.h>
#include <limits.h>
#include <SDL/SDL.h>

#define BUFSZ PIPE_BUF

#define BPP 0
#define CHILD_RES_X 200
#define CHILD_RES_Y 200
#define PARENT_RES_X 160
#define PARENT_RES_Y 40

/* First, let's define our colors */
enum {
	RED, BLUE, GREEN, YELLOW, TOTAL_COLORS
};

int fd[2]; /* File descriptor array for the pipe */
int pid;

void err_quit(char *msg);
void child(void);
void parent(void);

int main(void)
{
	/* Create the pipe we will use */
	if((pipe(fd)) < 0)
		err_quit("pipe");
		
	/* Fork and close appropriate descriptors */
	if((pid = fork()) < 0)
		err_quit("fork");
	
	if(pid == 0) {
		/* We are in child */
		child();
	} else {
		/* We are in parent */
		parent();
	}
	
	/* Wait for exit status */
	waitpid(pid, NULL, 0);
	
	exit(EXIT_SUCCESS);
}

void child(void)
{
	/* Child process... display a window, wait for instructions from parent as what
	  color to fill the window with */
	SDL_Surface *screen;
	char buf[BUFSZ];
	int len, i;
	Uint32 video_flags, lib_flags, color[TOTAL_COLORS];
	
	lib_flags = SDL_INIT_VIDEO;
	video_flags = (SDL_HWPALETTE | SDL_DOUBLEBUF);
	
	close(fd[1]); /* Close the write descriptor */
	if (SDL_Init (lib_flags) < 0) {
		fprintf (stderr, "Couldn't initialize SDL: %s\n", SDL_GetError ());
		exit(EXIT_FAILURE);
	}
	screen = SDL_SetVideoMode (CHILD_RES_X, CHILD_RES_Y, BPP, video_flags);
	/* No Cursor in child */
	SDL_ShowCursor(0);
	SDL_WM_SetCaption ("Watch Here", "multiwin");
	
	/* Define our colors */
	color[RED] = SDL_MapRGB (screen->format, 200, 0, 0);
	color[BLUE] = SDL_MapRGB (screen->format, 0, 0, 200);
	color[GREEN] = SDL_MapRGB (screen->format, 0, 200, 0);
	color[YELLOW] = SDL_MapRGB (screen->format, 200, 200, 0);
	
	/* Read the input from parent. If the input is one of the enumerated
	  colors, then fill the child window with that color */
	while((len = read(fd[0], buf, BUFSZ)) > 0) {
		i = *buf - *"0"; /* Turn the char back into an int */
		if((i >= 0) & (i < TOTAL_COLORS)) {
			SDL_FillRect(screen, NULL, color[i]);
			SDL_UpdateRect (screen, 0, 0, 0, 0);
		}
	}
	
	close(fd[0]); /* Close the file descriptor, we're done */
	
	SDL_FreeSurface(screen);
}

void parent(void)
{
	/* Parent process.. display a window with multi-color spheres in it, poll for clicks
	  and send color change requests to the child */
	
	SDL_Surface *screen;
	SDL_Surface *sphere[TOTAL_COLORS];
	SDL_Rect dest;
	SDL_Event event;
	char buf[BUFSZ];
	int i;
	Uint32 video_flags, lib_flags;
	
	lib_flags = SDL_INIT_VIDEO;
	video_flags = (SDL_HWPALETTE | SDL_DOUBLEBUF);
	
	close(fd[0]); /* Close the read descriptor */
	if (SDL_Init (lib_flags) < 0) {
		fprintf (stderr, "Couldn't initialize SDL: %s\n", SDL_GetError ());
		exit(EXIT_FAILURE);
	}
	screen = SDL_SetVideoMode (PARENT_RES_X, PARENT_RES_Y, BPP, video_flags);
	/* Want a cursor in Parent */
	SDL_ShowCursor(1);
	SDL_WM_SetCaption ("Chick Here", "multiwin");
	
	sphere[RED] = SDL_LoadBMP("red.bmp");
	sphere[BLUE] = SDL_LoadBMP("blue.bmp");
	sphere[GREEN] = SDL_LoadBMP("green.bmp");
	sphere[YELLOW] = SDL_LoadBMP("yellow.bmp");
	
	/* Bitmaps assumed to be of identical size */
	dest.y = 0;
	dest.w = sphere[RED]->w;
	dest.h = sphere[RED]->h;
	for(i = 0; i < TOTAL_COLORS; i++) {
		dest.x = i * dest.w;
		SDL_BlitSurface(sphere[i], NULL, screen, &dest);
	}
	
	SDL_UpdateRect (screen, 0, 0, 0, 0);
	
	/* Check on the user input in the parent window. If the user
	  has selected one of the color spheres, then send the
	  enumerated color to the child */
	while (SDL_PollEvent(&event) >= 0)
	{
		/* Determine if a sphere has been clicked */
		if (event.type == SDL_MOUSEBUTTONDOWN) {
			int x, y;
			SDL_GetMouseState (&x, &y);
			
			for(i = 0; i < TOTAL_COLORS; i++) {
				dest.x = i * dest.w;
				if ((x > dest.x) & (x < (dest.x + dest.w))) {
					/* Clicked within a color sphere, send a char
					  of the enumerated color to the child */
					sprintf(buf, "%i\n", i);
					write(fd[1], buf, strlen(buf));
				}
			}
		}

		if (event.type & (SDL_KEYUP | SDL_KEYDOWN))
		{
			Uint8 *keys = SDL_GetKeyState(&i);
			/* Press ESC to exit */
			if(keys[SDLK_ESCAPE])
				break;
		}
	}
	
	close(fd[1]); /* Close the file descriptor, we're done */
	
	SDL_FreeSurface(screen);
	for(i = 0; i < TOTAL_COLORS; i++)
		SDL_FreeSurface(sphere[i]);
}

void err_quit(char *msg)
{
	perror(msg);
	exit(EXIT_FAILURE);
}